/* =======================
     low_level functions
   ======================= */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "config.h"
/* #define MEMDEBUG */

/* ryk ==> */
#ifdef ZMC3_STATIC_LIB	/* Shared lib */
	#include "etc.h"
	#include "setjmp.h"
	#include "switch.h"
	jmp_buf	g_zmc3mark;
	char*	g_zmc3msgp = NULL;
	int		g_zmc3msglen = 0;
#endif
/* <== ryk */

#define	FILENOTFOUND	1
#define	NONZMD			2
#define	ZMDVERERR		3
#define	READERR			4
#define	WRITEERR		5
#define	FLUSHERR		6
#define	SWERR			7
#define	TOOMANYFILES	8
#define	WAVNOTDEFIND	9
#define	NOMEMORY		10
#define	ILLEGALSWITCH	11
#define	ILLEGALMEMORY	12
#define	COMPILEERROR	13

/* ryk ==> */
#ifndef ZMC3_STATIC_LIB	/* Shared lib */
typedef struct memDebugData {
	struct memDebugData *next;
	/* struct memDebugData *prev; */
	DWORD	size;		/* malloced size */
	void	*adrs;		/* malloced area */
	char	*label;		/* label name */
} MEMDEBUGDATA;
#endif
/* <== ryk */

#ifdef MEMDEBUG
void memdebug(void);
#endif
MEMDEBUGDATA *searchHimself(void *adrs, char *label, char f);
void initMemMaster(void);

void efreeAll(void);
void efreeAll2(char *label);
void efree(void *ptr, char *label);
void *emalloc(size_t n, char *label);
void *erealloc(void *ptr,size_t n, char *label);
void *emarealloc(void *ptr,size_t n, char *label);
void fatal(int ret,char *fmt, ...);
void terminateProgram(int);
inline WORD  getWordAlign(UBYTE*);
inline DWORD getDwordAlign(UBYTE*);
inline void  putWordAlign(UBYTE*,WORD);
inline void  putDwordAlign(UBYTE*,DWORD);
inline WORD  getWord(UBYTE*);
inline DWORD getDword(UBYTE*);
inline void  putWord(UBYTE*,WORD);
inline void  putDword(UBYTE*,DWORD);
#ifndef UNIXY_OS
int iskanji(int a);
#endif


#ifndef UNIXY_OS
int iskanji(int a)			/* only SJIS is supported. */
{
	return ((0x81 <= a && a <= 0x9F) || (0xE0 <= a && a <= 0xFC))? 1 : 0;
}
#endif



#ifdef UNIXY_OS
	/* ---- for use of `ntohs' ,`ntohl', `htons', and `htonl' ---- */
	#ifdef	FreeBSD
		#include <sys/param.h>
	#else
		#include <sys/types.h>
		#include <netinet/in.h>
	#endif
#endif	/* UNIXY_OS */



#ifdef A
static MEMDEBUGDATA memMaster = {
	NULL,	/* next */
	0,		/* size */
	NULL,	/* adrs */
	NULL	/* label */
};
#endif
MEMDEBUGDATA *memMaster = NULL;
#ifdef MEMDEBUG
static int nn=0,cc=0;
#endif

void initMemMaster(void)
{
	memMaster = (MEMDEBUGDATA*)malloc(sizeof(MEMDEBUGDATA) * 1);
	if (memMaster==(MEMDEBUGDATA*)NULL) {
		fprintf(stderr, "memMaster initialize failed.\n");
		exit(ILLEGALMEMORY);
	}
	memMaster->next = NULL;
	memMaster->size = 0;
	memMaster->adrs = NULL;
	memMaster->label = NULL;
/* fprintf(stderr,"[%p]\n",memMaster); */
}

/* #ifdef MEMDEBUG */
void memdebug(void)
{
	MEMDEBUGDATA *m = memMaster->next;
#ifdef MEMDEBUG
	int c = 0;
#endif

	while (m != NULL) {
#ifdef MEMDEBUG
		fprintf(stderr,"%p %d %s/%p\n",m->adrs, m->size, m->label, m->next);
		c++;
#endif
		m = m->next;
	}
#ifdef MEMDEBUG
	if (c) {
		fprintf(stderr, "%d memory blocks are still allocated.\n", c);
	}
#endif
}
/* #endif */


void efreeAll(void)
{
	MEMDEBUGDATA *m;
	int c = 0;

	if (memMaster) {
		m = memMaster->next;

		while (m) {
			MEMDEBUGDATA *n = m->next;
#ifdef MEMDEBUG
			fprintf(stderr,"%p %d %s\n",m->adrs, m->size, m->label);
#endif
			efree(m->adrs, m->label);
			m = n;
			c++;
		}
		free(memMaster);
#ifdef MEMDEBUG
		if (c) {
			fprintf(stderr, "%d memory blocks are freed.\n", c);
		}
#endif
		memMaster = NULL;
	}
#ifdef MEMDEBUG
if (nn) {
	fprintf(stderr,"[%d,%d,%d]",nn,cc,cc/nn);
}
#endif
}


void efreeAll2(char *label)
{
	MEMDEBUGDATA *m, *n;

/* fprintf(stderr,"efreeAll2: %s\n",label); */
	if (memMaster) {
		m = memMaster;
		n = m->next;

		while (n) {
/* fprintf(stderr,"a[%s]",n->label); */
			if (!strcmp(n->label,label)) {
/* fprintf(stderr,"0/%p/",n); */
				m->next = n->next;
				free(n->label);
				free(n->adrs);
				free(n);
				n = m;
				/* n = m; */
			}
			m = n;
			if (n) {
/* fprintf(stderr,"!"); */
				n = n->next;
/* fprintf(stderr,"8\n"); */
			}
		}
	}
/* fprintf(stderr,"9"); */
}


void efree(void *ptr, char *label)
{
#ifdef MEMDEBUG
nn++;
#endif
	if (ptr) {
		MEMDEBUGDATA *p, *m = NULL, *n = NULL;

		p = searchHimself(ptr, label, 'F');

		if (p) {
			if (p->next) {
				m = p->next;
			}
			if (m && m->next) {
				n = m->next;
			}

			free(m->label);
			free(m);
			p->next = n;

			free(ptr);
		}
	}
}

/* =======================================
     malloc function with error handling
   ======================================= */

void *emalloc(size_t n, char *label)
{
	void *pt;
	MEMDEBUGDATA *m;

	if ( (pt = malloc(n)) == NULL ) {
		fatal(NOMEMORY,"malloc failed.\n");
	}

	if ( (m = malloc(sizeof(MEMDEBUGDATA) * 1)) == NULL) {
		free(pt);
		fatal(NOMEMORY,"malloc failed.\n");
	}

	m->next = memMaster->next;
	memMaster->next = m;
	m->size = n;
	m->adrs = pt;
#ifdef MEMDEBUG
if (pt == NULL) {
	fprintf(stderr,"!!!MALLOC NULL!!!\n");
}
#endif
	m->label = malloc(sizeof(char) * strlen(label) + 1);
	if (m->label == NULL) {
		fatal(NOMEMORY,"malloc failed.\n");
	}
	strcpy(m->label, label);
#ifdef MEMDEBUG
/* fprintf(stderr,"M: %p %d %s\n", pt, n, label); */
#endif
	return pt;
}




/* ========================================
     search malloced table before
     If it found himself, return the "previous" pointer
     or NULL
   ======================================== */
/* #ifdef MEMDEBUG */
MEMDEBUGDATA *searchHimself(void *adrs, char *label, char f)
{
	MEMDEBUGDATA *m = memMaster;
	MEMDEBUGDATA *p = NULL;		/* previous adrs */
	MEMDEBUGDATA *r = NULL;		/* return value */

	if (!memMaster) {
		return NULL;
	}

	while (m->next != NULL) {
		p = m;
		m = m->next;
#ifdef MEMDEBUG
if (f=='F')
cc++;
#endif
		if (m->adrs == adrs) {
			r = p;
			break;
		}
	}

	if (r == NULL) {
/*		memdebug();
*/
		fatal (ILLEGALMEMORY,"MEMORY LEAK FOUND: mallocated table search failed.\n(%p [%c]%s)\n",adrs,f,label);
	}
	return r;
}
/* #endif */

/* ========================================
     realloc function with error handling
   ======================================== */

void *erealloc(void *ptr, size_t n, char *label)
{
	void *pt;
	MEMDEBUGDATA *m = searchHimself(ptr, label, 'r');

	if ( (pt = realloc(ptr,n)) == NULL ) {
#ifdef MEMDEBUG
		/* memdebug(); */
#endif
		fprintf(stderr,"%s: %p -> %d",label, ptr,n);
		fatal(NOMEMORY,"realloc failed.\n");
	}

	m = m->next;
	m->size = n;
	m->adrs = pt;
#ifdef MEMDEBUG
if (pt == NULL) {
	fprintf(stderr,"!!!REALLOC NULL!!!\n");
}
#endif
	m->label = realloc(m->label, strlen(m->label) + strlen(label) + 1);
	if (m->label == NULL) {
		fatal(NOMEMORY,"realloc failed.\n");
	}
	strcat(m->label, label);
#ifdef MEMDEBUG
/* fprintf(stderr,"R: %p %d %s\n", pt, n, label); */
#endif
	return pt;
}


void *emarealloc(void *ptr, size_t n, char *label)
{
	if ( !ptr ) {
		return emalloc(n,label);
	} else {
		return erealloc(ptr,n,label);
	}
}


/* ========================
     fatal error
   ======================== */

void fatal(int ret,char *fmt, ...)
{
	va_list ap;


	va_start(ap,fmt);
	vfprintf(stderr,fmt,ap);
	va_end(ap);

	terminateProgram(ret);
}


/* ===============================================
     program termination with cleaning-up memory
   =============================================== */

void terminateProgram(int ret)
{
	efreeAll();
	exit(ret);
}


/* =========================
	 read WORD,DWORD data
	 big/little endian OK
		for even-byte align only
		input:pointer
		output:number
   ========================= */

inline WORD getWordAlign(UBYTE *zmd)
{
#ifdef X68000
	return *(WORD*)zmd;
#else
	WORD data;

	data  = *zmd++ << 8;
	data += *zmd;
	return  data;
#endif
}

inline DWORD getDwordAlign(UBYTE *zmd)
{
#ifdef X68000
	return *(DWORD*)zmd;
#else
	int data;

	data  = (*zmd++ << 24);
	data += (*zmd++ << 16);
	data += (*zmd++ <<  8);
	data +=  *zmd;
	return data;
#endif
}

/* ============================
	 write WORD,DWORD data
	 big/little endian OK
		for even-byte align only
		input:number,pointer
   ============================ */

inline void putWordAlign(UBYTE *pos,WORD data)
{
#ifdef UNIXY_OS
	WORD	val;
	UBYTE	*ptr;

	val = htons(data);
	ptr = (UBYTE*)&val;

	*pos++ = *ptr++;
	*pos++ = *ptr++;
#elif defined X68000
	*(WORD*)pos = data;
#else
	*pos++ = data >> 8;
	*pos   = (data & 0xFF);
#endif
}


inline void putDwordAlign(UBYTE *pos,DWORD data)
{
#ifdef UNIXY_OS
	DWORD	val;
	UBYTE	*ptr;

	val = htonl(data);
	ptr = (UBYTE*)&val;

	*pos++ = *ptr++;
	*pos++ = *ptr++;
	*pos++ = *ptr++;
	*pos++ = *ptr++;
#elif defined X68000
	*(DWORD*)pos = data;
#else
	*pos++ =  data >> 24;
	*pos++ = (data >> 16) & 0xFF;
	*pos++ = (data >>  8) & 0xFF;
	*pos   = (data & 0xFF);
#endif
}


/* ============================
	 read WORD,DWORD data
		input:pointer
		output:number
   ============================ */

inline WORD getWord(UBYTE *zmd)
{
	WORD data;

	data  = *zmd++ << 8;
	data |= *zmd;
	return  data;
}

inline DWORD getDword(UBYTE *zmd)
{
	DWORD data;

	data  = (*zmd++ << 24);
	data |= (*zmd++ << 16);
	data |= (*zmd++ <<  8);
	data |=  *zmd;
	return data;
}

/* ============================
	 write WORD,DWORD data
		input:number,pointer
   ============================ */

inline void putWord(UBYTE *pos,WORD data)
{
	*pos++ = data >> 8;
	*pos   = (data & 0xFF);
}

inline void putDword(UBYTE *pos,DWORD data)
{
	*pos++ =  data >> 24;
	*pos++ = (data >> 16) & 0xFF;
	*pos++ = (data >>  8) & 0xFF;
	*pos   = (data & 0xFF);
}

/* ryk ==> */
/* =============================
	 Buffering error messages
   ============================= */
#ifdef ZMC3_STATIC_LIB	/* Shared lib */
static WORD swtbl_org[26];
void ZMC3CreateMessageBuf(void)
{
	extern void resetError();
	
	g_zmc3msgp = malloc(MAXERRMSG);
	g_zmc3msglen = 0;
	
	if(g_zmc3msgp != NULL) {
		*g_zmc3msgp = '\0';
	}

	resetError();
}
#endif
/* ============================
	 Add error messages
   ============================ */
void ZMC3FormatMessage(FILE* fp, const char* fmt, ...)
{
	va_list ap;
	unsigned char msg[256];
	va_start(ap,fmt);
	vsprintf(msg,fmt,ap);
	va_end(ap);

	fputs(msg, fp);
}

void ZMC3AddMessage(FILE* fp, const char* pmsg)
{
#ifdef ZMC3_STATIC_LIB	/* Static lib */
	int len;
	if(g_zmc3msgp) {
		len = strlen(pmsg);
		if(g_zmc3msglen + len > (MAXERRMSG - 37)) {
			if(g_zmc3msglen < MAXERRMSG)
			strcat(g_zmc3msgp, "--- Overflow message buffer ---\n");
			g_zmc3msglen = MAXERRMSG;
		} else {
			strcat(g_zmc3msgp, pmsg);
			g_zmc3msglen += len;
		}
	}
#else
	fputs(pmsg, fp);
#endif
}
/* ============================
	 Get error messages
   ============================ */
#ifdef ZMC3_STATIC_LIB	/* Shared lib */
char* ZMC3GetMessageBuf(void)
{
	return(g_zmc3msgp);
}
#endif

/* ============================
	 Dispose error messages
   ============================ */
	
#ifdef ZMC3_STATIC_LIB	/* Shared lib */
void ZMC3DisposeMessageBuf(void)
{
	if(g_zmc3msgp) {
		free(g_zmc3msgp);
	}
	g_zmc3msglen = 0;
}
#endif
/* <== ryk */
