/* ==================================
     Roland SC-88 Pro MML
   ================================== */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "parsesub.h"
#include "makezmd1.h"
#include "etc.h"

#include "structs.h"
#include "structs2.h"



UBYTE *makeZmdSc88ProEfxSwitch(
		UBYTE *zms,TRKCHINF *trkdata, const BYTE target[], int mode);
UBYTE *makeZmdSc88ModeSet(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[]);
UBYTE *makeZmdSc88ProEfxEqualizer(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[]);
UBYTE *makeZmdGsPartSetup(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[],
							 COMMONINF *cominf, const int cmd);

UBYTE *getIdIfCh(UBYTE *zms, TRKCHINF *trkdata, int Trk, int idflag,
					DWORD *id, DWORD *ifno, DWORD *ch);
UBYTE *macroExclusive88Pro(UBYTE *zms, BYTE e[], const DWORD elen,
							const int pos, const char *cmnt,
							int idflag, int roland, TRKCHINF *trkdata,
							const BYTE target[]);
int channel2Block(int ch);



extern DWORD line;
extern char *linebuf;
extern LINEDATA *ld;



/* ====================================================================*/
/* SC-88 Pro MML */
/* ====================================================================*/

UBYTE *makeZmdSc88ProEfxSwitch(
		UBYTE *zms,TRKCHINF *trkdata, const BYTE target[], int mode)
/*
.sc88Pro_efx_on id,if {ch}
.sc88Pro_efx_off id,if {ch}
.sc88_efx_bypass id,if {ch}
[sc88Pro_efx_on id,if {ch}]
[sc88Pro_efx_off id,if {ch}]
[sc88_efx_bypass id,if {ch}]

40 4x 22 00 / 0:BYPASS(off) / 01:EFX(on)
*/
/* mode: 0=off(bypass), 1=on(efx) */
{
	const DWORD elen = 10;
	BYTE e[10] = { 0xf0,0x41,0x10,0x42,0x12,0x40,0x40,0x22,0x00,0x00 };
	int pos = 6;
	const int rolandchecksum = 1;
	const int idflag = 1;

	e[9] = mode;
	zms = macroExclusive88Pro(zms, e, elen, pos,
								"SC88PRO_EFX_SWITCH",
								idflag, rolandchecksum, trkdata, target);

	zmserror(".sc88pro_efx_* can use only for zmc3.",linebuf,zms,4,1);


	return zms;
}


UBYTE *makeZmdSc88ModeSet(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[])
			/* 41 id 42 12 00 00 7f xx */
{
	int trk;
	UBYTE *zms_ = zms;

	for (trk = 0; target[trk] >= 0; trk++) {
		DWORD mode,id,ifno = -1;
		int err;
		const int elen = 9;
		BYTE e[9] = { 0xf0,0x41,0x10,0x42,0x12,0x00,0x00,0x7f,0x00 };
		const int Trk = target[trk];
		const char *p[] = {"single", "double",NULL};
		const char *SC88_MODE = "SC88_MODE";

		zms = getnum4(zms_,&mode,&err, p, 1);
		if (err) {
			zmserror("SYNTAX ERROR",linebuf,zms,0,1);
		} else if (mode < 1 || 2 < mode) {
			zmserror(".SC88_MODE: the mode is out of range.",linebuf,zms,0,1);
		}
		e[8] = mode - 1;

		zms = skipchr(zms,", \t\x0d\x0a");
		zms = getnum2(zms,&id,&err);
		if (err < 0) {
			id = 0x10;
		} else if (err > 0) {
			zmserror("SYNTAX ERROR",linebuf,zms,0,1);
		} else if (id < 0) {
			/* It is not to be error in V2, but.... */
		}
		e[2] = id;

		if (Trk == 0) {		/* common trk */
			zms = skipchr(zms,", \t\x0d\x0a");
			zms = getnum2(zms,&ifno,&err);
			if (err < 0) {
				ifno = 0xFF;
			} else if (err > 0) {
				zmserror("SYNTAX ERROR",linebuf,zms,0,1);
			} else if (ifno < 0) {
				/* It is not to be error in V2, but.... */
			}
		}

		*trkdata[Trk].zmd++ = (Trk)? 0xF3 : 0x34;
		*trkdata[Trk].zmd++ = (Trk)? 0x41 : ifno;
		*trkdata[Trk].zmd++ = strlen(SC88_MODE);
		strncpy(trkdata[Trk].zmd, SC88_MODE, strlen(SC88_MODE));
		trkdata[Trk].zmd += strlen(SC88_MODE);
		if (Trk) {
			putDword(trkdata[Trk].zmd, 4);
			trkdata[Trk].zmd += 4;
			memcpy(trkdata[Trk].zmd, &e[5], elen - 5);
			trkdata[Trk].zmd += elen - 5;
		} else {
			putDword(trkdata[Trk].zmd, (elen + 2) );	/* +2: checksum,0xf7 */
			trkdata[Trk].zmd += 4;
			trkdata[Trk].zmd = setex(trkdata[Trk].zmd, e, elen, 1);
			*trkdata[Trk].zmd++ = 0xf7;
		}
	}

	return zms;
}


UBYTE *makeZmdSc88ProEfxEqualizer(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[])
/* .SC88PRO_EFX_EQUALIZER id,if on/off */
/* $F0,$41,$10,$42,$12,$40,$03,$1F,0/1,sum,$F7 */
{
	DWORD dev, ifno = -1, trk;
	int err;
	const int f0f7 = 1,roland = 1;	/* Roland with checksum */
	const UBYTE mdl = 0x41;
	int		elen = 3;
	UBYTE e[3] = {0x40,0x03,0x1F};
	/* DWORD zmsskip; */

	zms = getnum2(zms,&dev,&err);
	if (err > 0) {
		zmserror("SYNTAX ERROR",linebuf,zms,0,1);
	} else if (err < 0) {
		dev = 0x10;
	}

	zms = skipSpc(zms);
	if (*zms == ',') {
		zms = getnum2(++zms, &ifno,&err);
		if (err < 0) {
			ifno = -1;
		} else if (err > 0) {
			zmserror("SYNTAX ERROR",linebuf,zms,0,1);
		} else if (ifno < 0) {
			/* It is not to be error in V2 */
		}
	}

	for (trk = 0; target[trk] >= 0; trk++) {
		const int Trk = target[trk];
		zms = exclusive(zms, f0f7, ifno, "SC88PRO_EFX_EQUALIZER_SWITCH",
									roland,dev,mdl,
									elen,e,
									"\"","\"\x0d\x0a",
									trkdata, Trk);
	}

	/* cominf->z_cmn_flag |= (1 << (0x34 / 4)); */

	return zms;
}


UBYTE *makeZmdGsPartSetup(UBYTE *zms,TRKCHINF *trkdata, const BYTE target[],
							 COMMONINF *cominf, const int cmd)
{	/* cmd: SC55=0 SC88=1 */
	const UBYTE f0f7 = 1,roland = 1;
	const UBYTE mdl= 0x42;
	DWORD dev, ifno = -1;
	int err, trk;
	DWORD part;		/* , zmsskip; */
	const DWORD elen = 3;
	BYTE e[3] = { 0x40,0x10,0x02 };
	/* int paras = (cmd == 1)? 127 : 119; */
	char *cmnt[] = {"GS_PART_SETUP","SC88_PART_SETUP"};

	zms = getnum2(zms,&part,&err);
	if (err) {
		zmserror("SYNTAX ERROR",linebuf,zms,0,1);
	}
	zms = getnum2(zms,&dev,&err);
	if (err == -1) {
		dev = 0x10;
	} else if (err > 0) {
		zmserror("SYNTAX ERROR",linebuf,zms,0,1);
	} else if (dev < 0) {
		/* It is not to be error in V2, but.... */
	}
	if (part == 10) {
		part = 0;
	} else if (part > 10) {
		part--;
	}
	e[1] = 0x10 + part;
	zms = skipSpc(zms);
	if (*zms == ',') {
		zms = getnum2(++zms, &ifno,&err);
		if (err < 0) {
			ifno = -1;
		} else if (err > 0) {
			zmserror("SYNTAX ERROR",linebuf,zms,0,1);
		} else if (ifno < 0) {
			/* It is not to be error in V2 */
		}
	}
	for (trk = 0; target[trk] >= 0; trk++) {
		const int Trk = target[trk];
		UBYTE *zmd_ = trkdata[Trk].zmd;
		const int offset = 3 + strlen(cmnt[cmd]) + 4 + 8;
		UBYTE *psum = zmd_ + offset;

		zms = exclusive(zms, f0f7, ifno, cmnt[cmd],
						roland,dev,mdl,
						elen,e,
						"{","}, \t\x0d\x0a",
						trkdata, Trk);

		while (*++psum != 0xF7) {
		}
		psum--;		/* check sum */


		if ( (*(zmd_ + offset))-- == 0) {	/* rewrite Rx channel # */
			*(zmd_ + offset) = 0x10;
			*psum -= 15;
		} else {
			(*psum)++;
		}
		*psum &= 0x7F;
#ifdef AAA
		if (trkdata[0].zmd - zmd_ > 13 + paras) {		/* non FULL parameters */
			/* 18 ll ll f0 41 10 42 12 40 1x 02 n1 n2 ... x127 sum f7 */
			if (paras == 119) {
				zmserror(".sc55_part_setup: too many parameters.",linebuf,zms,2,1);
			} else {
				zmserror(".sc88_part_setup: too many parameters.",linebuf,zms,3,1);
			}
		}
#endif
		cominf->z_cmn_flag |= (1 << (0x34 / 4));
	}

	return zms;
}





/* ====================================================================*/
/* useful subroutines */
/* ====================================================================*/

UBYTE *getIdIfCh(UBYTE *zms, TRKCHINF *trkdata, int Trk, int idflag,
					DWORD *id, DWORD *ifno, DWORD *ch)
{
	int err;

	zms = getIdIf(zms, idflag, id, ifno);

	zms = skipSpc(zms);
	if (*zms == '{') {
		zms = getnum2(++zms,ch,&err);
		if (err == -1) {
			*ch = (Trk)? trkdata[Trk].ch : -1;
		} else if (err > 0) {
			zmserror("channel number error.",linebuf,zms,0,1);
		}
		zms = skipSpc(zms);
		if (*zms != '}') {
			zmserror("SYNTAX ERROR.",linebuf,zms,0,1);
		}
	}

	return zms;
}


UBYTE *macroExclusive88Pro(UBYTE *zms, BYTE e[], const DWORD elen,
							const int pos, const char *cmnt,
							int idflag, int roland, TRKCHINF *trkdata,
							const BYTE target[])
{
	int trk;

	for (trk = 0; target[trk] >= 0; trk++) {
		int err, Trk = target[trk];
		DWORD id, ifno = 0xFF, ch;

		zms = getIdIfCh(zms, trkdata, Trk, idflag, &id, &ifno, &ch);
		if (ch == -1) {
			zmserror("channel number is required.",linebuf,zms,0,1);
		}
		e[2] = id;

		e[pos] |= channel2Block(ch);

		macroExclusiveMain(e, elen, cmnt, ifno, roland, trkdata, Trk);
	}
	return zms;
}


int channel2Block(int ch)	/* convert from channel# to block# */
{
	if (ch == 10) {
		ch = 0;
	} else if (ch > 10) {
		ch--;
	}

	return (ch & 0xF);
}
