/* ========================================
	making hash table for the program name
   ======================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "etc.h"


#include "structs2.h"
#include "config.h"
#ifndef	EXIT_SUCCESS
	#define	EXIT_SUCCESS	0
#endif
#ifndef	HASHTABLESIZE
	#define	HASHTABLESIZE	0x2000
#endif

void initHash(PROGHASH ***tbl);
void freeHash(PROGHASH **tbl);
WORD calcHash(char *str, int mode);
WORD calcHash2(PROGHASH *tbl[], char *str, DWORD *depth2);
void setHashTable(PROGHASH *tbl[], const UWORD hash,
		const char *name, const UBYTE pc, const BYTE msb, const BYTE lsb);
PROGHASH *refHash(PROGHASH *tbl[], UDWORD hashno);



void initHash(PROGHASH ***tbl)
{
	int i;

	*tbl = (PROGHASH**)emalloc(sizeof(PROGHASH*) * HASHTABLESIZE, "Hash table");

	for (i = 0; i < HASHTABLESIZE; i++) {
		(*tbl)[i] = (PROGHASH*)NULL;
	}
}


void freeHash(PROGHASH **tbl)
{
	int i;
	for (i = 0; i < HASHTABLESIZE; i++) {
		if (tbl[i] != NULL) {
			efree(tbl[i]->progname, "progname");
			efree(tbl[i], "Hash entry");
		}
	}
	efree(tbl, "Hash table");
}


WORD calcHash(char *str, int mode)
{
	int				i;
	unsigned int	p = 0;
	WORD			sum = 0;

	if (mode == 0) {
		while (p < strlen(str)) {
			UWORD	sum2 = 0;
			for (i = 0; i < 2; i++) {
				sum2 <<= 8;
				if (str[p]) {
					sum2 += str[p++];
				}
			}
			sum += sum2;
			sum &= 0x7fff;
		}
	} else {
		while (p < strlen(str)) {
			sum += str[p++];
			sum &= 0x7fff;
		}
	}

	sum %= HASHTABLESIZE;

	return sum;
}




/* To calcHash until the collision is avoided */
WORD calcHash2(PROGHASH *tbl[], char *str, DWORD *depth2)
{
	WORD	hash = calcHash(str, 0);
	WORD	uniqflg = 1;
/* printf("%.4x: %d/%d/%d/%s\n",sum,pc,msb,lsb,prg); */

	while (tbl[hash] != (PROGHASH*)NULL) {
		const	UWORD	hash2 = calcHash(str, 1);
/*
printf("Col: %.4x : %d/%d/%d/%s / %d/%d/%d/%s\n", hash,
tbl[hash]->pc, tbl[hash]->msb, tbl[hash]->lsb, tbl[hash]->progname,
pc, msb, lsb, prg);
*/
		hash += hash2;
		hash %= HASHTABLESIZE;
		(*depth2)++;

		if (tbl[hash] != (PROGHASH*)NULL && !strcmp(tbl[hash]->progname,str)) {
			uniqflg = 0;
		}
	}

	if (!uniqflg) {
		hash = -1;
	}
	return hash;
}



void setHashTable(PROGHASH *tbl[], const UWORD hash,
		const char *name, const UBYTE pc, const BYTE msb, const BYTE lsb)
{
	tbl[hash]			= (PROGHASH*)emalloc(sizeof(PROGHASH) * 1,
							"makehash entry");
	tbl[hash]->progname	= (char*)emalloc(sizeof(char) * (strlen(name) + 1),
							"progname");
	tbl[hash]->hash		= hash;
	strcpy(tbl[hash]->progname, name);
	tbl[hash]->pc		= pc;
	tbl[hash]->msb		= msb;
	tbl[hash]->lsb		= lsb;
}


PROGHASH *refHash(PROGHASH *tbl[], UDWORD hashno)
{
	if (hashno < 0 || HASHTABLESIZE <= hashno) {
		return (PROGHASH*)NULL;
	}

	return tbl[hashno];
}
