/* ================
     switch
   ================ */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "etc.h"
#include "help.h"

#define SET_VAL	-2
#define RES_VAL	-1
int getSwitch2(char *sw);
int getSwitch(int argc,char **argv);
void redirectErrorMessagesOpen(int mode, char *filename);
int getSwitchVal(char letter);
void setSwitchVal(char letter,char num);
char *getInFile(void);
char *getOutFile(void);
char *setInFile(char *filename);
char *setOutFile(char *filename);
void initInOutFile(void);
void freeInOutFile(void);
void initSwitch(void);
void releaseSwitch(void);

#ifndef WARNLVL
	#define WARNLVL	3
#endif



const int getnum(char **ptr);
void swerr(void);
#include "selstmod.h"

/* ryk ==> */
static WORD swtbl_org[26] =
						 {	RES_VAL,RES_VAL,RES_VAL,RES_VAL,RES_VAL,
						/*  A       B       C       D      E        */
							RES_VAL,RES_VAL,RES_VAL,RES_VAL,RES_VAL,
						/*  F       G       H       I      J        */
							RES_VAL,1,      RES_VAL,RES_VAL,RES_VAL,
						/*  K       L       M       N      O        */
							RES_VAL,RES_VAL,RES_VAL,RES_VAL,RES_VAL,
						/*  P       Q       R       S      T        */
							RES_VAL,3      ,WARNLVL,RES_VAL,RES_VAL,1};
						/*  U       V       W       X      Y        Z */
static WORD *swtbl;
static char *infile = NULL, *outfile = NULL;
static char NULLSTR[1] = { '\0' };
static FILE *stderr_orig;


/* ==============================
     get command line emuration
     convert from

      argv = "-x6 -i1 -z3" style to

      argv[0] = "-x6"
      argv[1] = "-i1"
      argv[2] = "-z3"
      argc = 3   style
   ============================== */

int getSwitch2(char *sw)
{
	int argc = 1, quot = 0, c = 0;
	char *p = sw;
	char **argv;

	while (*p) {
		if (isspace(*p) && !quot) {
			argc++;
		}
		if (*p == '\"' && ( ( p == sw) || *(p - 1) != '\\')) {
			quot = 1 - quot;
		}
		p++;
	}

	if (quot) {
		fatal(SWERR,"Error: quotation is not closed in the switch strings.\n");
	}

	argv = (char**)emalloc(sizeof(char*) * argc, "getSwitch2_argv**");
	p = sw;
	{
		char *p2 = p;

		while (*p2) {
			if (isspace(*p2) && !quot) {
				argv[c] = (char*)emalloc(sizeof(char) * (p2 - p + 1), "getSwitch2_argv*");
				strncpy(argv[c], p, p2 - p);
				argv[c][p2 - p] = '\0';
				c++;
				p = p2 + 1;		/* cut space */
			}
			if (*p2 == '\"' && (( p2 == sw) || *(p2 - 1) != '\\')) {
				quot = 1 - quot;
			}
			p2++;
		}

		if (!quot) {
			argv[c] = (char*)emalloc(sizeof(char) * (p2 - p + 1), "getSwitch2_argv*");
			strncpy(argv[c], p, p2 - p);
			argv[c][p2 - p] = '\0';
		}
	}

	getSwitch(argc, argv);
	for (c = 0; c < argc; c++) {
		efree(argv[c], "getSwitch2_argv*");
	}
	efree(argv, "getSwitch2_argv");

	return 0;
}

/* ================================
     get command line
     before coming here,
     call "initInOutFile()" first
     to allocate memory area for filenames
   ================================ */

int getSwitch(int argc,char **argv)
{
	infile[0] = outfile[0] = '\0';

	if (argc == 1) {

#ifdef UNIXY_OS
		strcpy(infile,"-");
		setSwitchVal('c',SET_VAL);
#else
		return 1;
#endif
	}

	while (--argc > 0) {
		char *opt, *opt_;

		opt_ = opt = (char *)emalloc(sizeof(char) * (strlen(*++argv) + 1), "optstr");
		strcpy( opt,*argv );

		while (*opt == ' ' || *opt == '\t') {
			opt++;
		}
#ifdef UNIXY_OS
		if ( opt[0]=='-' ) {
#else
		if ( opt[0]=='-' || opt[0]=='/' ) {
#endif
			/* char *ptr = strlwr(opt); */
			*opt = tolower(*++opt);
			if (!*opt) {						/* use standard input */
				if (!infile[0]) {
					infile[0] = '-';
					infile[1] = '\0';
				} else {
					fatal(TOOMANYFILES,"Error: too many sources are specified.\n");
				}
			}
			while (*opt) {
				char ch = *opt;
				if (ch == '?') {
					help(0,0);
				} else if (!isalpha(ch)) {
					fatal(SWERR,"Error: %c: it's an invalid switch.\n",ch);
				} else {
					if (ch == 'o') {			/* destination filename will be specified */
						if (*++opt) {
							setOutFile(opt);
							while (*opt) opt++;
						} else if (--argc) {
							setOutFile(*++argv);
							/* argv++; */
							/* argc--; */
						} else {
							fatal(SWERR,"Error: -o: no ZMD file specified.\n");
						}
						break;
					} else if (ch == 'e') {
						/* ERROR filename will be specified */
						if (*++opt) {
							redirectErrorMessagesOpen(0, opt);
							while (*opt) opt++;
						} else if (--argc) {
							redirectErrorMessagesOpen(0, *++argv);
						} else {
							fatal(SWERR,"Error: -e: no output file specified.\n");
						}
						break;
					} else if ( isdigit((int)*++opt) ) {
						setSwitchVal(ch, getnum(&opt));
					} else {
						setSwitchVal(ch, SET_VAL);
					}
				}
			}
		} else {
			if (!infile[0]) {
				setInFile(opt);
			} else {
				fatal(TOOMANYFILES,"Error: too many sources are specified.\n" );
			}
		}
		efree(opt_, "optstr");
	}
	return 0;
}



void redirectErrorMessagesOpen(int mode, char *filename)
		/* mode == 0: 1st to redirect (initially "w" open)
		 *         1: 2nd to redirect ("w+" open)
		 *         2: reset the redirect to stderr
		 */
{
	FILE *fp;
	char *modestr[2] = { "w", "w+" };

	if (mode < 2) {
		stderr_orig = stderr;
		fp = freopen(filename, modestr[mode], stderr);
		if (fp == (FILE*)NULL) {
			fatal(SWERR,"Error: -e: failed to redirect.\n");
		}
	} else {
		fclose(stderr);
		/* stderr = stderr_orig; */
	}
}

/* ==============================
     pick up numerics
   ============================== */

const int getnum(char **ptr)
{
	int n = 0;

	if (!**ptr) {
		swerr();
	}
	while ( isdigit((int)**ptr) ) {
		n = n * 10 + (**ptr) - '0';
		(*ptr)++;
	}
	if (n < 0) swerr();
	return n;
}


/* ==================================
     numeric err 
   ================================== */

void swerr(void)
{
	fprintf( stderr,"invalid number.\n" );
	help(1,0);
}


/* ===================================
     return specified switch's value
   =================================== */
   
int getSwitchVal(char letter)
{
	letter |= 0x20;

	return swtbl[ letter - 'a' ];
}

void setSwitchVal(char letter,char num)
{
	letter |= 0x20;

	swtbl[ letter - 'a' ] = num;
}



char *getInFile(void)
{
	return infile? infile : NULLSTR;
}

char *getOutFile(void)
{
	return outfile? outfile : NULLSTR;
}

char *setInFile(char *filename)
{
	infile = (char *)erealloc(infile, sizeof(char) * (strlen(filename) + 1), "infile_area");
	return strcpy(infile,filename);
}
char *setOutFile(char *filename)
{
	outfile = (char *)erealloc(outfile, sizeof(char) * (strlen(filename) + 1), "outfile_area");
	return strcpy(outfile,filename);
}

void initInOutFile(void)
{
	infile = (char *)emalloc(sizeof(char) * 2049, "infile_area");
	outfile = (char *)emalloc(sizeof(char) * 2049, "outfile_area");
	initSwitch();
}

void freeInOutFile(void)
{
	releaseSwitch();
	efree(outfile, "outfile_area");
	efree(infile, "infile_area");
}

void initSwitch(void)
{
	/* if (!swtbl) { */	/* ryk */
		swtbl = (WORD*)emalloc(sizeof(WORD) * 26, "SwitchTable");
	/* } */	/* ryk */
	memcpy(swtbl, swtbl_org, sizeof(WORD) * 26);
}

void releaseSwitch(void)
{
	efree(swtbl,"SwitchTable");
	swtbl = NULL;	/* ryk */
}
