/* ======================================
     ZMC3.{X,EXE} ... ZMS->ZMD compiler
     for non X680x0 architectures
   ====================================== */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "config.h"

#include "selstmod.h"

#include "switch.h"
#include "readfile.h"
#include "makezmd.h"
#include "etc.h"
#include "help.h"

/* ryk ==> */
#ifdef ZMC3_STATIC_LIB		/* Static lib */
	#include "setjmp.h"
	extern jmp_buf g_zmc3mark;
#endif
/* <== ryk */

UBYTE *convertZmstoZmd(char *sw, UBYTE *zmsbuf, DWORD *zmdsize);
UBYTE *convertZmstoZmdMain(UBYTE *zmsbuf, DWORD *zmdsize);
FILE *openOutFile(void);
UBYTE *makeZmdFile(TRKCHINF *trkdata, COMMONINF *cominf, DWORD totalstep,
					DWORD *filesize);
int writeZmdFile(UBYTE *zbuf, DWORD filesize);

void checkSwitch(void);
UBYTE *makeZmdHeaderBlock(TRKCHINF *trkdata,COMMONINF *cominf, DWORD totalstep);
UBYTE *makeTrkChInfoTable(TRKCHINF *trkdata,COMMONINF *cominf, UDWORD *bytes);
void convertCR(char *p);

#ifndef EXIT_SUCCESS
	#define EXIT_SUCCESS	0
#endif

#define TOTALHEAD	80		/* header size */

int main(int argc,char *argv[])
{
	UBYTE		*zmsbuf, *zmdbuf;
	DWORD		zmdsize;
#ifdef UNIXY_OS
	const char *inpEXT[] = { ".zms", ".zm3", ".opm", NULL};
	const char *outEXT  = ".zmd";
#else
	const char *inpEXT[] = { ".ZMS", ".ZM3", ".OPM", NULL};
	const char *outEXT  = ".ZMD";
#endif

/* ryk ==> */
#ifdef ZMC3_STATIC_LIB	/* Static lib */
	int			result;
	extern		jmp_buf g_zmc3mark;

	result = setjmp(g_zmc3mark);
	if(result != 0) {
		return(result);
	}
#endif
/* <== ryk */


	fprintf(stderr,"ZMS to ZMD(V3) compiler version 0.97, by Mamiya / YokoMuMu / Y.Yagi,\n");
	fprintf(stderr,"from Z-MUSIC UNIVERSAL VERSION 3.02C (C) ZENJI SOFT\n");

	initMemMaster();							/* initialize emalloc/efree entry */
	srand(time(NULL));

	initInOutFile();
	/* setvbuf(stderr,NULL,_IONBF,0); */		/* no output buffering */
	if (getSwitch(argc,argv)) {					/* get switches/sourcefilename/dest.file */
		help(0,0);								/* bootup without any arguments */
	}
	checkSwitch();

	if (!getInFile()[0]) {
#ifdef UNIXY_OS
		setInFile("-");
#else
		fatal(FILENOTFOUND,"Error: source file is not specified.\n");
#endif
	}

	zmsbuf = readFile(getInFile(), getOutFile(), inpEXT, outEXT);

#ifdef TEST
	zmsbuf = malloc(8192);
	{
		FILE *fp;
		int z;
		fp = fopen("test.zms","rb");
		z = fread(zmsbuf, sizeof(UBYTE), 8192, fp);
		zmsbuf[z++] = 0xff;
		zmsbuf[z++] = 0xff;
		zmsbuf[z++] = 0xff;
		zmsbuf[z++] = 0xff;
		fclose(fp);
	}
zmdbuf = convertZmstoZmd("-x6 -i1", zmsbuf, &zmdsize);
#endif
	zmdbuf = convertZmstoZmdMain(zmsbuf, &zmdsize);
	writeZmdFile(zmdbuf, zmdsize);

	/* efree(zmdbuf, "OUTPUT BIN DATA"); */
	free(zmdbuf);		/* zmdbuf must be free-ed in the normal way */

	freeInOutFile();

	efree(zmsbuf, "zms_in_v3macro");

#ifdef MEMDEBUG
	/* memdebug(); */
#endif
	efreeAll();				/* to clear memMaster used in etc.c */

	return	EXIT_SUCCESS;
}



/* =================================================
     the convertion entry for the another programs
       input: 1) the pointer to the "switches"
                (e.g. "-x6 -i1" (without quote))
              2) the pointer to the zms
                (gurdian '0xFF 0xFF' are needed
                 at the tail of zms data!!!)
              3) the pointer ZMD size should be
                 copied (it may used for your debug)
      output: the pointer to the zmd

       you have to free() the "zmd" memory by yourself
   ================================================= */

UBYTE *convertZmstoZmd(char *sw, UBYTE *zmsbuf, DWORD *zmdsize)
{
	UBYTE	*zmdbuf;

	initMemMaster();				/* initialize emalloc/efree entry */
	srand(time(NULL));

	initInOutFile();
	getSwitch2(sw);
	checkSwitch();
	zmdbuf = convertZmstoZmdMain(zmsbuf, zmdsize);
	freeInOutFile();
	efreeAll();

	return zmdbuf;
}



UBYTE *convertZmstoZmdMain(UBYTE *zmsbuf, DWORD *zmdsize)
{
	UBYTE		*zmdbuf;
	TRKCHINF	*trkdata;
	COMMONINF	cominf;
	DWORD totalstep = 0;

	convertCR(zmsbuf);							/* convert CR code */
	trkdata = makeZmd(zmsbuf, getInFile(), &cominf, &totalstep);
	zmdbuf = makeZmdFile(trkdata, &cominf, totalstep, zmdsize);

	return zmdbuf;
}



/* =======================================================
     open output file (with dispatching FILE and STDOUT)
   ======================================================= */

FILE *openOutFile(void)
{
	int stdoutflg = 0;
	FILE *fp;

	if (getSwitchVal('c') != RES_VAL) {		/* use stdout */
		stdoutflg = 1;
	}

	if (stdoutflg) {
		#include "stdinbin.h"
		fp = stdout;
	} else {
		fp = fopen(getOutFile(),"wb");
		if (fp == (FILE*)NULL) {
			fatal(FILENOTFOUND, "%s\tError: destination file can't open.",getOutFile());
		}
	}
	return fp;
}


/* =======================
     write all ZMD files
   ======================= */

int writeZmdFile(UBYTE *zbuf, DWORD filesize)
{
	FILE *fp;
	UDWORD bytes;

	fp = openOutFile();

	bytes = fwrite(zbuf, sizeof(UBYTE), filesize, fp);
	if (fp != stdout && bytes < (UDWORD)filesize) {
		fatal(WRITEERR, "%s\tError: destination file can't write.",getOutFile());
	}

	if (fp != stdout) {
		fclose(fp);
	}

	return 0;
}



/* =======================
     make all ZMD files
   ======================= */

UBYTE *makeZmdFile(TRKCHINF *trkdata, COMMONINF *cominf, DWORD totalstep,
					DWORD *filesize)
{
#ifdef Aa
	int trks = 0;
#endif
	int i, firsttrk = 1;
	UDWORD bytes2, outsize;
	UBYTE *headdata, *trktbl;
	UBYTE *outdata_, *outdata;
	DWORD outpos;

	/* header block */
	headdata = makeZmdHeaderBlock(trkdata, cominf, totalstep);
	outdata_ = (UBYTE*)emalloc(sizeof(UBYTE) * TOTALHEAD, "OUTPUT BIN DATA");
	outpos = outsize = TOTALHEAD;
	memcpy(outdata_, headdata, TOTALHEAD);
	efree(headdata,"headdata");


/* common track */
	if (trkdata[0].nowsize) {				/* if there is a common track */ 
		const DWORD s = trkdata[0].nowsize;
		outsize += s;
		outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
		memcpy(outdata_ + outpos, trkdata[0].zmdbuf, s);
		outpos += s;
	}
	efree(trkdata[0].zmdbuf,"zmdbuf");

/* trk/ch info.table */
	trktbl = makeTrkChInfoTable(trkdata, cominf, &bytes2);
	outsize += bytes2;
	outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
	memcpy(outdata_ + outpos, trktbl, bytes2);
	outpos += bytes2;
	efree(trktbl,"trktbl");


	/* trk data / trk additional info */
#ifdef AAA
	for (i = 1; cominf.trkassign[i] >= 0; i++) {
		if (trkdata[Trk].mute && (*trkdata[Trk].zmdbuf != 0xFF || getSwitchVal('z') >= 3)) {
			trks++;
		}
	}
/* printf("trks=%d\n",trks); */
	putWordAlign(tmp,trks);
	bytes = fwrite(tmp,sizeof(UBYTE),2,fp);
	if (!stdoutflg && bytes < 2) {
		fatal(WRITEERR, "%s\tError: destination file can't write.",getOutFile());
	}
	offset = trks * 6 - 4;
	/* hedder part2 */
	for (i = 0; i < trks; i++) {
		/* putDwordAlign(tmp,offset); */
		putDword(tmp,offset);
		*(tmp + 4) = 0;
		*(tmp + 5) = trkdata[cominf.trkassign[i]].ch;
		bytes = fwrite(tmp,sizeof(UBYTE),6,fp);
		if (!stdoutflg && bytes < 6) {
			fatal(WRITEERR, "%s\tError: destination file can't write.",getOutFile());
		}

		offset += trkdata[cominf.trkassign[i]].nowsize;
		offset -= 6;
	}

	/* data */
#endif






	for (i = 0; cominf->trkassign[i] >= 0; i++) {
		const int Trk = cominf->trkassign[i];
		if ((trkdata[Trk].mute && *trkdata[Trk].zmdbuf != 0xFF) || getSwitchVal('z') >= 3) {
			if (firsttrk) {				/* +2 padding for TRKS(word) */
				outsize += 2;
				outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
				*(outdata_ + outpos++) = 0x2F;
				*(outdata_ + outpos++) = 0x2F;
				firsttrk = 0;
			}
			outsize += 16;
			outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
			putDword(outdata_ + outpos    , 0);	/* track total step time */
			putDword(outdata_ + outpos + 4, 0);	/* track checksum */
			putDword(outdata_ + outpos + 8, 0);	/* pars */
			if (trkdata[Trk].cmnt) {			/* comment length */
				putDword(outdata_ + outpos + 12, strlen(trkdata[Trk].cmnt));
			} else {
				putDword(outdata_ + outpos + 12, 0);
			}
			outpos += 16;

			if (trkdata[Trk].cmnt) {
				const DWORD s = strlen(trkdata[Trk].cmnt);
				outsize += s;
				outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
				memcpy (outdata_ + outpos, trkdata[Trk].cmnt, s);
				outpos += s;
				if (s & 1) {						/* padding if need */
					outsize++;
					outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
					*(outdata_ + outpos++) = 0;
				}
			}
#ifdef A
		if (len & 3) {						/* padding if need */
			int i;
			for (i = (len & 3); i < 4; i++) {
				fputc(0, fp);
			}
		}
#endif

			{
				const DWORD s = trkdata[Trk].nowsize;
				outsize += s;
				outdata_ = (UBYTE*)erealloc(outdata_, sizeof(UBYTE) * outsize, "OUTPUT BIN DATA");
				memcpy(outdata_ + outpos, trkdata[Trk].zmdbuf, s);
				outpos += s;
			}
#ifdef Aa
{
	FILE *fp2;
	char file[256];

	sprintf(file,"%d.tmp",Trk);
	fp2 = fopen(file,"wb");
	fwrite(trkdata[Trk].zmdbuf,sizeof(UBYTE),
							trkdata[Trk].nowsize,fp2);
	fclose(fp2);
}
#endif

		}

		if (trkdata[Trk].cmnt) {
			efree(trkdata[Trk].cmnt,"trkdata[Trk].cmnt");
		}
		efree(trkdata[Trk].zmdbuf, "trkdata[Trk].zmdbuf");
	}


#ifdef AA
	bytes = fwrite(outdata_, sizeof(UBYTE), outsize, fp);
	if (!stdoutflg && bytes < outsize) {
		fatal(WRITEERR, "%s\tError: destination file can't write.",getOutFile());
	}
/*
	efree(outdata_, "OUTPUT BIN DATA");
*/

	if (!stdoutflg) {
		fclose(fp);
	}
#endif
/*
	efree(trktbl);
	efree(headdata);
*/
	efree(trkdata,"trkdata");


	outdata = (UBYTE*)malloc(sizeof(UBYTE) * outsize);
	if (!outdata) {
		fatal(NOMEMORY, "malloca failed.");
	}
	memcpy(outdata, outdata_, outsize);
	efree(outdata_, "OUTPUT BIN DATA");

	*filesize = outsize;
	return outdata;
}





/* ==================
     check switches
   ================== */

void checkSwitch()
{
	int tmp;

	tmp = getSwitchVal('h');
	if (tmp == 1) {
		help(0,1);
	} else if (tmp != RES_VAL) {
		help(0,0);
	}

	if (getOutFile()[0] && getSwitchVal('c') != RES_VAL) {
		fatal(TOOMANYFILES,"You cannot use both -c and -o.\n");
	}
	if (getSwitchVal('w') == SET_VAL) {
		fatal(SWERR,"You cannot omit -w's parameter.\n");
	}
}


/* =========================================================
     make ZMD header table
     this table is generated even if
     there are no common track and no playng track.
   ========================================================= */

UBYTE *makeZmdHeaderBlock(TRKCHINF *trkdata,COMMONINF *cominf, DWORD totalstep)
{
	UBYTE *head, *head_;
	int i;
	/* DWORD trks = 0; */
	DWORD totaltrksize = 2 + 2;		/* 2(trks.w) and 2(padding) */
	int usedch[16 * 4 + 8 + 16] = {	/* MIDI + FM + ADPCM */
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0
	};


	head_ = head = (UBYTE*)emalloc(sizeof(UBYTE) * TOTALHEAD, "headerBlock");

	strcpy(head,"\x1aZmuSiC\x30");	/* Z-MUSIC ID */
	head += 8;

	if (trkdata[0].nowsize) {		/* offset to CommonBlock */ 
		putDword(head, TOTALHEAD - (head - head_) - 4);
	} else {
		putDword(head, 0);
	}
	head += 4;
/*
fprintf(stderr,"total=%d ",totaltrksize);
*/



	for (i = 0; cominf->trkassign[i] >= 0; i++) {
		const int Trk = cominf->trkassign[i];
		int offset = 0;

		if (trkdata[Trk].chtype == 0) {			/* FM */
			offset = 16 * 4;
		} else if (trkdata[Trk].chtype == 1) {	/* ADPCM */
			offset = 16 * 4 + 8;
		} else if (0x8000 <= trkdata[Trk].chtype && trkdata[Trk].chtype < 0x8004) {
			offset = (trkdata[Trk].chtype - 0x8000) * 16;	/* MIDI */
		} else if (trkdata[Trk].chtype == 0xFFFF) {	/* current MIDI */
			offset = cominf->currentmidiout * 16;
		}
		usedch[offset + trkdata[Trk].ch]++;
	}


	for (i = 0; cominf->trkassign[i] >= 0; i++) {
		const int Trk = cominf->trkassign[i];
		if ((trkdata[Trk].mute && *trkdata[Trk].zmdbuf != 0xFF) || (getSwitchVal('z') >= 3) ) {
			/* trks++; */
			totaltrksize += trkdata[Trk].nowsize;
/*
fprintf(stderr,"total=%d ",totaltrksize);
*/
										/* playing trk */
			totaltrksize += 16;			/* size for trk/ch info. table */
			totaltrksize += 16;			/* trk additional info */
			if (trkdata[Trk].cmnt) {
				totaltrksize += strlen(trkdata[Trk].cmnt);
			}
			while (totaltrksize & 3) {
				totaltrksize++;
			}
		}
	}
#ifdef A
	if (trkdata[0].zmdbuf[0] == 0xFF) {		/* if there is no data in commontrk */
		totaltrksize -= 2;		/* delete 0xFF 0xFF */
	}
#endif

/* fprintf(stderr,"trks=%x total=%x ",trks,totaltrksize); */


	if (totaltrksize) {				/* offset to trk/ch info. */
		putDword(head, TOTALHEAD - (head - head_) - 4 + trkdata[0].nowsize);
	} else {
		putDword(head, 0);
	}
/*
fprintf(stderr,"TOTALHEAD=%x, t0=%x, total=%x\n",TOTALHEAD,trkdata[0].nowsize, totaltrksize);
*/

/*========*/
	head += 4;
	putDword(head, 0);				/* 10: no support for playcommand */
	head += 4;
	putDword(head, TOTALHEAD + trkdata[0].nowsize + totaltrksize);
									/* 14: ZMD file size */
	head += 4;
	putDword(head, 0);				/* 18: no support for WRD data */
	head += 4;
	putDword(head, 0);				/* 1C: reserved */
	head += 4;
/*========*/
	putDword(head, 0);				/* 20: total step time offset */
	head += 4;
	if (cominf->cmnt) {				/* 24: offset to the title */
		putDword(head, TOTALHEAD - 0x28 + cominf->cmnt);
	} else {
		putDword(head, 0);
	}
	head += 4;
	putDword(head, totalstep);		/* 28: total step time */
	head += 4;
	putDword(head, 0);				/* 2C: playing time */
	head += 4;
/*========*/
	putWord(head, cominf->meter);	/* 30: meter n/m */
	head += 2;
	*head++ = 0;					/* 32: metronome speed */
	*head++ = 0;					/* 33: reserved */
	*head++ = cominf->key;			/* 34: key */
	*head++ = cominf->majorflag;	/* 35: major / minor */
	putWord(head, cominf->div);		/* 36: master clock */
	head += 2;
	putWord(head, cominf->tempo);	/* 38: tempo */
	head += 2;
	putWord(head, 0);				/* 3A: reserved */
	head += 2;
	putDword(head, cominf->z_cmn_flag);	/* 3C: com trk flag */
	head += 4;
/*=========*/
	putDword(head, 0);				/* 40: playcommand flag */
	head += 4;
	putDword(head, 0);				/* 44: the type of instruments */
	head += 4;

	*head = 0;						/* #of FM ch */
	for (i = 16 * 4; i < 16 * 4 + 8; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	*++head = 0;					/* #of ADPCM ch */
	for (i = 16 * 4 + 8; i < 16 * 4 + 8 + 16; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	*++head = 0;					/* #of MIDI-1 ch */
	for (i = 16 * 0 + 0; i < 16 * 0 + 16; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	*++head = 0;					/* #of MIDI-2 ch */
	for (i = 16 * 1 + 0; i < 16 * 1 + 16; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	*++head = 0;					/* #of MIDI-3 ch */
	for (i = 16 * 2 + 0; i < 16 * 2 + 16; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	*++head = 0;					/* reserved (#of MIDI-4 ch) */
	for (i = 16 * 3 + 0; i < 16 * 3 + 16; i++) {
		if (usedch[i]) (*head) += usedch[i];
	}
	head++;
	putWord(head, 0);				/* reserved */

	return head_;
}


UBYTE *makeTrkChInfoTable(TRKCHINF *trkdata,COMMONINF *cominf, UDWORD *bytes)
{
	int i;
	DWORD trks = 0, trk;
	UBYTE *buf, *buf_;
	DWORD headsize = 2;
	DWORD trkofst;
	int firsttrk = 1;

	for (i = 0; cominf->trkassign[i] >= 0; i++) {
		int Trk = cominf->trkassign[i];
		if ((trkdata[Trk].mute && *trkdata[Trk].zmdbuf != 0xFF) || getSwitchVal('z') >= 3) {
			trks++;
			headsize += 16;				/* size for trk/ch info. table */
		}
	}
	/* *bytes = 2 + trks * 16; */
	trkofst = headsize - 2 - 8 - 8;		/* */

/*fprintf(stderr,"[size: %x]\n",headsize);*/
	buf_ = buf = (UBYTE*)emalloc(sizeof(UBYTE) * headsize, "headerCommon");
	putWord(buf, trks - 1);
	buf += 2;

/*	for (trk = 0; trk < trks; trk++) {*/
	for (trk = 0; cominf->trkassign[trk] >= 0; trk++) {
		int Trk = cominf->trkassign[trk];
		if ((trkdata[Trk].mute && *trkdata[Trk].zmdbuf != 0xFF) || getSwitchVal('z') >= 3 ) {

			*buf++ = 0;							/* performance track */
			if (trkdata[Trk].trkm) {			/* track mode: normal/rhythm */
				*buf++ = 0x80;
			} else {
				*buf++ = 0;
			}
			*buf++ = trkdata[Trk].trkf - 1;		/* freq */
/* fprintf(stderr,"freq=%d ",trkdata[Trk].trkf - 1); */
			*buf++ = 0;							/* reserved */
			putWord(buf, trkdata[Trk].chtype);	/* channel */
/* fprintf(stderr,"chtype=%d ",trkdata[Trk].chtype); */
			buf += 2;
			putWord(buf, trkdata[Trk].ch_);		/* channel */
/* fprintf(stderr,"ch=%d ",trkdata[Trk].ch_); */
			buf += 2;

			buf += 4;							/* offset to track additional info. */

			/* if (trk == 0) { */
			if (firsttrk) {
				trkofst += 2;					/* +2 for trks[word] */
				firsttrk = 0;
			}

/*fprintf(stderr,"adinfo=%x ",trkofst);*/
			putDword(buf, trkofst);
			trkofst += 16;
			if (trkdata[Trk].cmnt) {
				trkofst += strlen(trkdata[Trk].cmnt);
/*fprintf(stderr,"(%d: %s)",Trk,trkdata[Trk].cmnt);*/
			}
			trkofst += 2;
			while (trkofst & 1) {
				trkofst++;						/* padding if need */
			}
#ifdef AA
			if (trk == 0) {						/* +2 for padding */
				trkofst += 2;
			}
#endif
			/* trkofst += 2; */
			trkofst += 2;
			buf -= 4;

			putDword(buf, trkofst);			/* offset to track data */

/*fprintf(stderr,"trkofst=%.2x ",trkofst );*/

			buf += 8;
			trkofst += trkdata[Trk].nowsize;
/*
fprintf(stderr,"[trkofst=%.2x] ",trkofst);
*/
			trkofst -= 16 + 4;
		}
	}

	*bytes = headsize;

	return buf_;
}
#ifdef AA
		totaltrksize += trkdata[i].nowsize;
		if (i) {						/* playing trk */
			totaltrksize += 18;			/* size for trk/ch info. table */
			if (trkdata[i].cmnt) {
				totaltrksize += 16 + strlen(trkdata[i].cmnt);
			}
		}
#endif


/* ===================
     convert CR code
   =================== */
void convertCR(char *p)
{
	--p;
	while (*++p != (char)0xFF) {		/* until gurdian-code comes... */
		if (*p == ignoreCR) {	/* ignoreCR: 0x0D for unix/DOS, 0x0A for Mac. */
			*p = SPACE;
		}
	}
}
